@@ -1,10 +1,14 @@
-use crate::logical_plan::{det_melt_schema, Context};
+#[cfg(feature = "ipc")]
+use crate::logical_plan::LpScanOptions;
+#[cfg(feature = "parquet")]
+use crate::logical_plan::ParquetOptions;
+use crate::logical_plan::{det_melt_schema, Context, CsvParserOptions};
 use crate::prelude::*;
 use crate::utils::{aexprs_to_schema, PushNode};
 use ahash::RandomState;
-use polars_core::frame::hash_join::JoinType;
+use polars_core::frame::groupby::DynamicGroupOptions;
 use polars_core::prelude::*;
-use polars_core::utils::{Arena, Node};
+use polars_utils::arena::{Arena, Node};
 use std::collections::HashSet;
 #[cfg(any(feature = "csv-file", feature = "parquet"))]
 use std::path::PathBuf;
@@ -22,7 +26,7 @@ pub enum ALogicalPlan {
     Slice {
         input: Node,
         offset: i64,
-        len: usize,
+        len: u32,
     },
     Selection {
         input: Node,
@@ -31,27 +35,34 @@ pub enum ALogicalPlan {
     #[cfg(feature = "csv-file")]
     CsvScan {
         path: PathBuf,
+        // schema of the complete file
         schema: SchemaRef,
-        has_header: bool,
-        delimiter: u8,
-        ignore_errors: bool,
-        skip_rows: usize,
-        stop_after_n_rows: Option<usize>,
-        with_columns: Option<Vec<String>>,
+        // schema of the projected file
+        output_schema: Option<SchemaRef>,
+        options: CsvParserOptions,
+        predicate: Option<Node>,
+        aggregate: Vec<Node>,
+    },
+    #[cfg(feature = "ipc")]
+    IpcScan {
+        path: PathBuf,
+        schema: SchemaRef,
+        // schema of the projected file
+        output_schema: Option<SchemaRef>,
+        options: LpScanOptions,
         predicate: Option<Node>,
         aggregate: Vec<Node>,
-        cache: bool,
-        low_memory: bool,
     },
     #[cfg(feature = "parquet")]
     ParquetScan {
         path: PathBuf,
+        // schema of the complete file
         schema: SchemaRef,
-        with_columns: Option<Vec<String>>,
+        // schema of the projected file
+        output_schema: Option<SchemaRef>,
         predicate: Option<Node>,
         aggregate: Vec<Node>,
-        stop_after_n_rows: Option<usize>,
-        cache: bool,
+        options: ParquetOptions,
     },
     DataFrameScan {
         df: Arc<DataFrame>,
@@ -87,16 +98,16 @@ pub enum ALogicalPlan {
         aggs: Vec<Node>,
         schema: SchemaRef,
         apply: Option<Arc<dyn DataFrameUdf>>,
+        maintain_order: bool,
+        dynamic_options: Option<DynamicGroupOptions>,
     },
     Join {
         input_left: Node,
         input_right: Node,
         schema: SchemaRef,
-        how: JoinType,
         left_on: Vec<Node>,
         right_on: Vec<Node>,
-        allow_par: bool,
-        force_par: bool,
+        options: JoinOptions,
     },
     HStack {
         input: Node,
@@ -117,6 +128,10 @@ pub enum ALogicalPlan {
         projection_pd: bool,
         schema: Option<SchemaRef>,
     },
+    Union {
+        inputs: Vec<Node>,
+        options: UnionOptions,
+    },
 }
 
 impl Default for ALogicalPlan {
@@ -131,18 +146,33 @@ impl Default for ALogicalPlan {
 }
 
 impl ALogicalPlan {
-    pub(crate) fn schema<'a>(&'a self, arena: &'a Arena<ALogicalPlan>) -> &'a Schema {
+    pub(crate) fn schema<'a>(&'a self, arena: &'a Arena<ALogicalPlan>) -> &'a SchemaRef {
         use ALogicalPlan::*;
         match self {
+            Union { inputs, .. } => arena.get(inputs[0]).schema(arena),
             Cache { input } => arena.get(*input).schema(arena),
             Sort { input, .. } => arena.get(*input).schema(arena),
             Explode { input, .. } => arena.get(*input).schema(arena),
             #[cfg(feature = "parquet")]
-            ParquetScan { schema, .. } => schema,
+            ParquetScan {
+                schema,
+                output_schema,
+                ..
+            } => output_schema.as_ref().unwrap_or(schema),
+            #[cfg(feature = "ipc")]
+            IpcScan {
+                schema,
+                output_schema,
+                ..
+            } => output_schema.as_ref().unwrap_or(schema),
             DataFrameScan { schema, .. } => schema,
             Selection { input, .. } => arena.get(*input).schema(arena),
             #[cfg(feature = "csv-file")]
-            CsvScan { schema, .. } => schema,
+            CsvScan {
+                schema,
+                output_schema,
+                ..
+            } => output_schema.as_ref().unwrap_or(schema),
             Projection { schema, .. } => schema,
             LocalProjection { schema, .. } => schema,
             Aggregate { schema, .. } => schema,
@@ -157,77 +187,6 @@ impl ALogicalPlan {
             },
         }
     }
-
-    /// Check ALogicalPlan equality. The nodes may differ.
-    ///
-    /// For instance: there can be two columns "foo" in the memory arena. These are equal,
-    /// but would have different node values.
-    #[cfg(feature = "private")]
-    pub(crate) fn eq(
-        node_left: Node,
-        node_right: Node,
-        lp_arena: &Arena<ALogicalPlan>,
-        expr_arena: &Arena<AExpr>,
-    ) -> bool {
-        use crate::logical_plan::iterator::ArenaLpIter;
-        use std::fs::canonicalize;
-
-        let cmp = |(node_left, node_right)| {
-            use ALogicalPlan::*;
-            match (lp_arena.get(node_left), lp_arena.get(node_right)) {
-                #[cfg(feature = "csv-file")]
-                (CsvScan { path: path_a, .. }, CsvScan { path: path_b, .. }) => {
-                    canonicalize(path_a).unwrap() == canonicalize(path_b).unwrap()
-                }
-                #[cfg(feature = "parquet")]
-                (ParquetScan { path: path_a, .. }, ParquetScan { path: path_b, .. }) => {
-                    canonicalize(path_a).unwrap() == canonicalize(path_b).unwrap()
-                }
-                (DataFrameScan { df: df_a, .. }, DataFrameScan { df: df_b, .. }) => {
-                    df_a.ptr_equal(df_b)
-                }
-                // the following don't affect the schema, but do affect the # of rows or the row order.
-                (Selection { predicate: l, .. }, Selection { predicate: r, .. }) => {
-                    AExpr::eq(*l, *r, expr_arena)
-                }
-                (
-                    Sort {
-                        by_column: l,
-                        reverse: r_l,
-                        ..
-                    },
-                    Sort {
-                        by_column: r,
-                        reverse: r_r,
-                        ..
-                    },
-                ) => l == r && r_l == r_r,
-                (Explode { columns: l, .. }, Explode { columns: r, .. }) => l == r,
-                (
-                    Distinct {
-                        maintain_order: l1,
-                        subset: l2,
-                        ..
-                    },
-                    Distinct {
-                        maintain_order: r1,
-                        subset: r2,
-                        ..
-                    },
-                ) => l1 == r1 && l2 == r2,
-                (a, b) => {
-                    std::mem::discriminant(a) == std::mem::discriminant(b)
-                        && a.schema(lp_arena) == b.schema(lp_arena)
-                }
-            }
-        };
-
-        lp_arena
-            .iter(node_left)
-            .zip(lp_arena.iter(node_right))
-            .map(|(tpll, tplr)| (tpll.0, tplr.0))
-            .all(cmp)
-    }
 }
 
 impl ALogicalPlan {
@@ -236,6 +195,10 @@ impl ALogicalPlan {
         use ALogicalPlan::*;
 
         match self {
+            Union { options, .. } => Union {
+                inputs,
+                options: *options,
+            },
             Melt {
                 id_vars,
                 value_vars,
@@ -270,6 +233,8 @@ impl ALogicalPlan {
                 keys,
                 schema,
                 apply,
+                maintain_order,
+                dynamic_options,
                 ..
             } => Aggregate {
                 input: inputs[0],
@@ -277,23 +242,21 @@ impl ALogicalPlan {
                 aggs: exprs[keys.len()..].to_vec(),
                 schema: schema.clone(),
                 apply: apply.clone(),
+                maintain_order: *maintain_order,
+                dynamic_options: dynamic_options.clone(),
             },
             Join {
                 schema,
-                how,
                 left_on,
-                allow_par,
-                force_par,
+                options,
                 ..
             } => Join {
                 input_left: inputs[0],
                 input_right: inputs[1],
                 schema: schema.clone(),
-                how: *how,
                 left_on: exprs[..left_on.len()].to_vec(),
                 right_on: exprs[left_on.len()..].to_vec(),
-                allow_par: *allow_par,
-                force_par: *force_par,
+                options: options.clone(),
             },
             Sort {
                 by_column, reverse, ..
@@ -321,14 +284,37 @@ impl ALogicalPlan {
                 exprs,
                 schema: schema.clone(),
             },
+            #[cfg(feature = "ipc")]
+            IpcScan {
+                path,
+                schema,
+                output_schema,
+                options,
+                predicate,
+                ..
+            } => {
+                let mut new_predicate = None;
+                if predicate.is_some() {
+                    new_predicate = exprs.pop()
+                }
+
+                IpcScan {
+                    path: path.clone(),
+                    schema: schema.clone(),
+                    output_schema: output_schema.clone(),
+                    predicate: new_predicate,
+                    aggregate: exprs,
+                    options: options.clone(),
+                }
+            }
+
             #[cfg(feature = "parquet")]
             ParquetScan {
                 path,
                 schema,
-                with_columns,
+                output_schema,
                 predicate,
-                stop_after_n_rows,
-                cache,
+                options,
                 ..
             } => {
                 let mut new_predicate = None;
@@ -339,26 +325,19 @@ impl ALogicalPlan {
                 ParquetScan {
                     path: path.clone(),
                     schema: schema.clone(),
-                    with_columns: with_columns.clone(),
+                    output_schema: output_schema.clone(),
                     predicate: new_predicate,
                     aggregate: exprs,
-                    stop_after_n_rows: *stop_after_n_rows,
-                    cache: *cache,
+                    options: options.clone(),
                 }
             }
             #[cfg(feature = "csv-file")]
             CsvScan {
                 path,
                 schema,
-                has_header,
-                delimiter,
-                ignore_errors,
-                skip_rows,
-                stop_after_n_rows,
-                with_columns,
+                output_schema,
                 predicate,
-                cache,
-                low_memory,
+                options,
                 ..
             } => {
                 let mut new_predicate = None;
@@ -368,16 +347,10 @@ impl ALogicalPlan {
                 CsvScan {
                     path: path.clone(),
                     schema: schema.clone(),
-                    has_header: *has_header,
-                    delimiter: *delimiter,
-                    ignore_errors: *ignore_errors,
-                    skip_rows: *skip_rows,
-                    stop_after_n_rows: *stop_after_n_rows,
-                    with_columns: with_columns.clone(),
+                    output_schema: output_schema.clone(),
+                    options: options.clone(),
                     predicate: new_predicate,
                     aggregate: exprs,
-                    cache: *cache,
-                    low_memory: *low_memory,
                 }
             }
             DataFrameScan {
@@ -428,6 +401,7 @@ impl ALogicalPlan {
             | Explode { .. }
             | Cache { .. }
             | Distinct { .. }
+            | Union { .. }
             | Udf { .. } => {}
             Selection { predicate, .. } => container.push(*predicate),
             Projection { expr, .. } => container.extend_from_slice(expr),
@@ -454,6 +428,17 @@ impl ALogicalPlan {
                     container.push(*node)
                 }
             }
+            #[cfg(feature = "ipc")]
+            IpcScan {
+                predicate,
+                aggregate,
+                ..
+            } => {
+                container.extend_from_slice(aggregate);
+                if let Some(node) = predicate {
+                    container.push(*node)
+                }
+            }
             #[cfg(feature = "csv-file")]
             CsvScan {
                 predicate,
@@ -489,13 +474,19 @@ impl ALogicalPlan {
 
     /// Push inputs of the LP in of this node to an existing container.
     /// Most plans have typically one input. A join has two and a scan (CsvScan)
-    /// or an in-memory DataFrame has none.
+    /// or an in-memory DataFrame has none. A Union has multiple.
     pub(crate) fn copy_inputs<T>(&self, container: &mut T)
     where
         T: PushNode,
     {
         use ALogicalPlan::*;
         let input = match self {
+            Union { inputs, .. } => {
+                for node in inputs {
+                    container.push_node(*node);
+                }
+                return;
+            }
             Melt { input, .. } => *input,
             Slice { input, .. } => *input,
             Selection { input, .. } => *input,
@@ -519,6 +510,8 @@ impl ALogicalPlan {
             Udf { input, .. } => *input,
             #[cfg(feature = "parquet")]
             ParquetScan { .. } => return,
+            #[cfg(feature = "ipc")]
+            IpcScan { .. } => return,
             #[cfg(feature = "csv-file")]
             CsvScan { .. } => return,
             DataFrameScan { .. } => return,
@@ -647,8 +640,10 @@ impl<'a> ALogicalPlanBuilder<'a> {
         keys: Vec<Node>,
         aggs: Vec<Node>,
         apply: Option<Arc<dyn DataFrameUdf>>,
+        maintain_order: bool,
+        dynamic_options: Option<DynamicGroupOptions>,
     ) -> Self {
-        debug_assert!(!keys.is_empty());
+        debug_assert!(!(keys.is_empty() && dynamic_options.is_none()));
         let current_schema = self.schema();
         // TODO! add this line if LogicalPlan is dropped in favor of ALogicalPlan
         // let aggs = rewrite_projections(aggs, current_schema);
@@ -665,6 +660,8 @@ impl<'a> ALogicalPlanBuilder<'a> {
             aggs,
             schema: Arc::new(schema),
             apply,
+            maintain_order,
+            dynamic_options,
         };
         let root = self.lp_arena.add(lp);
         Self::new(root, self.expr_arena, self.lp_arena)
@@ -673,17 +670,15 @@ impl<'a> ALogicalPlanBuilder<'a> {
     pub fn join(
         self,
         other: Node,
-        how: JoinType,
         left_on: Vec<Node>,
         right_on: Vec<Node>,
-        allow_par: bool,
-        force_par: bool,
+        options: JoinOptions,
     ) -> Self {
         let schema_left = self.schema();
         let schema_right = self.lp_arena.get(other).schema(self.lp_arena);
 
         // column names of left table
-        let mut names: HashSet<&String, RandomState> = HashSet::with_capacity_and_hasher(
+        let mut names: HashSet<&str, RandomState> = HashSet::with_capacity_and_hasher(
             schema_left.len() + schema_right.len(),
             Default::default(),
         );
@@ -691,7 +686,7 @@ impl<'a> ALogicalPlanBuilder<'a> {
         let mut fields = Vec::with_capacity(schema_left.len() + schema_right.len());
 
         for f in schema_left.fields() {
-            names.insert(f.name());
+            names.insert(f.name().as_ref());
             fields.push(f.clone());
         }
 
@@ -706,9 +701,9 @@ impl<'a> ALogicalPlanBuilder<'a> {
 
         for f in schema_right.fields() {
             let name = f.name();
-            if !right_names.contains(name) {
-                if names.contains(name) {
-                    let new_name = format!("{}_right", name);
+            if !right_names.contains(name.as_str()) {
+                if names.contains(name.as_str()) {
+                    let new_name = format!("{}{}", name, options.suffix.as_ref());
                     let field = Field::new(&new_name, f.data_type().clone());
                     fields.push(field)
                 } else {
@@ -722,12 +717,10 @@ impl<'a> ALogicalPlanBuilder<'a> {
         let lp = ALogicalPlan::Join {
             input_left: self.root,
             input_right: other,
-            how,
             schema,
             left_on,
             right_on,
-            allow_par,
-            force_par,
+            options,
         };
         let root = self.lp_arena.add(lp);
         Self::new(root, self.expr_arena, self.lp_arena)
