@@ -41,13 +41,13 @@ pub(crate) fn to_aexpr(expr: Expr, arena: &mut Arena<AExpr>) -> Node {
             expr: to_aexpr(*expr, arena),
             idx: to_aexpr(*idx, arena),
         },
-        Expr::Sort { expr, reverse } => AExpr::Sort {
+        Expr::Sort { expr, options } => AExpr::Sort {
             expr: to_aexpr(*expr, arena),
-            reverse,
+            options,
         },
         Expr::SortBy { expr, by, reverse } => AExpr::SortBy {
             expr: to_aexpr(*expr, arena),
-            by: to_aexpr(*by, arena),
+            by: by.into_iter().map(|e| to_aexpr(e, arena)).collect(),
             reverse,
         },
         Expr::Filter { input, by } => AExpr::Filter {
@@ -65,9 +65,14 @@ pub(crate) fn to_aexpr(expr: Expr, arena: &mut Arena<AExpr>) -> Node {
                 AggExpr::Mean(expr) => AAggExpr::Mean(to_aexpr(*expr, arena)),
                 AggExpr::List(expr) => AAggExpr::List(to_aexpr(*expr, arena)),
                 AggExpr::Count(expr) => AAggExpr::Count(to_aexpr(*expr, arena)),
-                AggExpr::Quantile { expr, quantile } => AAggExpr::Quantile {
+                AggExpr::Quantile {
+                    expr,
+                    quantile,
+                    interpol,
+                } => AAggExpr::Quantile {
                     expr: to_aexpr(*expr, arena),
                     quantile,
+                    interpol,
                 },
                 AggExpr::Sum(expr) => AAggExpr::Sum(to_aexpr(*expr, arena)),
                 AggExpr::Std(expr) => AAggExpr::Std(to_aexpr(*expr, arena)),
@@ -152,6 +157,13 @@ pub(crate) fn to_alp(
     lp_arena: &mut Arena<ALogicalPlan>,
 ) -> Node {
     let v = match lp {
+        LogicalPlan::Union { inputs, options } => {
+            let inputs = inputs
+                .into_iter()
+                .map(|lp| to_alp(lp, expr_arena, lp_arena))
+                .collect();
+            ALogicalPlan::Union { inputs, options }
+        }
         LogicalPlan::Selection { input, predicate } => {
             let i = to_alp(*input, expr_arena, lp_arena);
             let p = to_aexpr(predicate, expr_arena);
@@ -188,6 +200,7 @@ pub(crate) fn to_alp(
         } => ALogicalPlan::CsvScan {
             path,
             schema,
+            output_schema: None,
             options,
             predicate: predicate.map(|expr| to_aexpr(expr, expr_arena)),
             aggregate: aggregate
@@ -195,26 +208,41 @@ pub(crate) fn to_alp(
                 .map(|expr| to_aexpr(expr, expr_arena))
                 .collect(),
         },
-        #[cfg(feature = "parquet")]
-        LogicalPlan::ParquetScan {
+        #[cfg(feature = "ipc")]
+        LogicalPlan::IpcScan {
             path,
             schema,
-            with_columns,
             predicate,
             aggregate,
-            stop_after_n_rows,
-            cache,
-        } => ALogicalPlan::ParquetScan {
+            options,
+        } => ALogicalPlan::IpcScan {
             path,
             schema,
-            with_columns,
+            output_schema: None,
             predicate: predicate.map(|expr| to_aexpr(expr, expr_arena)),
             aggregate: aggregate
                 .into_iter()
                 .map(|expr| to_aexpr(expr, expr_arena))
                 .collect(),
-            stop_after_n_rows,
-            cache,
+            options,
+        },
+        #[cfg(feature = "parquet")]
+        LogicalPlan::ParquetScan {
+            path,
+            schema,
+            predicate,
+            aggregate,
+            options,
+        } => ALogicalPlan::ParquetScan {
+            path,
+            schema,
+            output_schema: None,
+            predicate: predicate.map(|expr| to_aexpr(expr, expr_arena)),
+            aggregate: aggregate
+                .into_iter()
+                .map(|expr| to_aexpr(expr, expr_arena))
+                .collect(),
+            options,
         },
         LogicalPlan::DataFrameScan {
             df,
@@ -285,6 +313,7 @@ pub(crate) fn to_alp(
             schema,
             apply,
             maintain_order,
+            dynamic_options,
         } => {
             let i = to_alp(*input, expr_arena, lp_arena);
             let aggs_new = aggs.into_iter().map(|x| to_aexpr(x, expr_arena)).collect();
@@ -300,6 +329,7 @@ pub(crate) fn to_alp(
                 schema,
                 apply,
                 maintain_order,
+                dynamic_options,
             }
         }
         LogicalPlan::Join {
@@ -423,11 +453,11 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
                 strict,
             }
         }
-        AExpr::Sort { expr, reverse } => {
+        AExpr::Sort { expr, options } => {
             let exp = node_to_exp(expr, expr_arena);
             Expr::Sort {
                 expr: Box::new(exp),
-                reverse,
+                options,
             }
         }
         AExpr::Take { expr, idx } => {
@@ -440,10 +470,13 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
         }
         AExpr::SortBy { expr, by, reverse } => {
             let expr = node_to_exp(expr, expr_arena);
-            let by = node_to_exp(by, expr_arena);
+            let by = by
+                .iter()
+                .map(|node| node_to_exp(*node, expr_arena))
+                .collect();
             Expr::SortBy {
                 expr: Box::new(expr),
-                by: Box::new(by),
+                by,
                 reverse,
             }
         }
@@ -489,11 +522,16 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
                 let exp = node_to_exp(expr, expr_arena);
                 AggExpr::List(Box::new(exp)).into()
             }
-            AAggExpr::Quantile { expr, quantile } => {
+            AAggExpr::Quantile {
+                expr,
+                quantile,
+                interpol,
+            } => {
                 let exp = node_to_exp(expr, expr_arena);
                 AggExpr::Quantile {
                     expr: Box::new(exp),
                     quantile,
+                    interpol,
                 }
                 .into()
             }
@@ -604,6 +642,13 @@ pub(crate) fn node_to_lp(
     let lp = std::mem::take(lp);
 
     match lp {
+        ALogicalPlan::Union { inputs, options } => {
+            let inputs = inputs
+                .into_iter()
+                .map(|node| node_to_lp(node, expr_arena, lp_arena))
+                .collect();
+            LogicalPlan::Union { inputs, options }
+        }
         ALogicalPlan::Slice { input, offset, len } => {
             let lp = node_to_lp(input, expr_arena, lp_arena);
             LogicalPlan::Slice {
@@ -624,6 +669,7 @@ pub(crate) fn node_to_lp(
         ALogicalPlan::CsvScan {
             path,
             schema,
+            output_schema: _,
             options,
             predicate,
             aggregate,
@@ -634,23 +680,35 @@ pub(crate) fn node_to_lp(
             predicate: predicate.map(|n| node_to_exp(n, expr_arena)),
             aggregate: nodes_to_exprs(&aggregate, expr_arena),
         },
+        #[cfg(feature = "ipc")]
+        ALogicalPlan::IpcScan {
+            path,
+            schema,
+            output_schema: _,
+            predicate,
+            aggregate,
+            options,
+        } => LogicalPlan::IpcScan {
+            path,
+            schema,
+            predicate: predicate.map(|n| node_to_exp(n, expr_arena)),
+            aggregate: nodes_to_exprs(&aggregate, expr_arena),
+            options,
+        },
         #[cfg(feature = "parquet")]
         ALogicalPlan::ParquetScan {
             path,
             schema,
-            with_columns,
+            output_schema: _,
             predicate,
             aggregate,
-            stop_after_n_rows,
-            cache,
+            options,
         } => LogicalPlan::ParquetScan {
             path,
             schema,
-            with_columns,
             predicate: predicate.map(|n| node_to_exp(n, expr_arena)),
             aggregate: nodes_to_exprs(&aggregate, expr_arena),
-            stop_after_n_rows,
-            cache,
+            options,
         },
         ALogicalPlan::DataFrameScan {
             df,
@@ -718,6 +776,7 @@ pub(crate) fn node_to_lp(
             schema,
             apply,
             maintain_order,
+            dynamic_options,
         } => {
             let i = node_to_lp(input, expr_arena, lp_arena);
 
@@ -728,6 +787,7 @@ pub(crate) fn node_to_lp(
                 schema,
                 apply,
                 maintain_order,
+                dynamic_options,
             }
         }
         ALogicalPlan::Join {
