@@ -1,6 +1,5 @@
 mod utils;
 
-use crate::logical_plan::optimizer::ALogicalPlanBuilder;
 use crate::logical_plan::{optimizer, Context};
 use crate::prelude::*;
 use crate::utils::{aexpr_to_root_names, aexprs_to_schema, check_input_node, has_aexpr};
@@ -12,7 +11,7 @@ use utils::*;
 pub(crate) struct PredicatePushDown {}
 
 impl PredicatePushDown {
-    fn apply_predicate(
+    fn optional_apply_predicate(
         &self,
         lp: ALogicalPlan,
         local_predicates: Vec<Node>,
@@ -29,27 +28,10 @@ impl PredicatePushDown {
         }
     }
 
-    fn finish_at_leaf(
-        &self,
-        lp: ALogicalPlan,
-        acc_predicates: PlHashMap<Arc<String>, Node>,
-        lp_arena: &mut Arena<ALogicalPlan>,
-        expr_arena: &mut Arena<AExpr>,
-    ) -> ALogicalPlan {
-        match acc_predicates.len() {
-            // No filter in the logical plan
-            0 => lp,
-            _ => {
-                let local_predicates = acc_predicates.into_iter().map(|t| t.1).collect();
-                self.apply_predicate(lp, local_predicates, lp_arena, expr_arena)
-            }
-        }
-    }
-
     fn pushdown_and_assign(
         &self,
         input: Node,
-        acc_predicates: PlHashMap<Arc<String>, Node>,
+        acc_predicates: PlHashMap<Arc<str>, Node>,
         lp_arena: &mut Arena<ALogicalPlan>,
         expr_arena: &mut Arena<AExpr>,
     ) -> Result<()> {
@@ -59,6 +41,106 @@ impl PredicatePushDown {
         Ok(())
     }
 
+    /// Filter will be pushed down.
+    fn pushdown_and_continue(
+        &self,
+        lp: ALogicalPlan,
+        mut acc_predicates: PlHashMap<Arc<str>, Node>,
+        lp_arena: &mut Arena<ALogicalPlan>,
+        expr_arena: &mut Arena<AExpr>,
+        has_projections: bool,
+    ) -> Result<ALogicalPlan> {
+        let inputs = lp.get_inputs();
+        let exprs = lp.get_exprs();
+
+        if has_projections {
+            // we should not pass these projections
+            if exprs
+                .iter()
+                .any(|e_n| is_pushdown_boundary(*e_n, expr_arena))
+            {
+                return self.no_pushdown_restart_opt(lp, acc_predicates, lp_arena, expr_arena);
+            }
+
+            // projections should only have a single input.
+            assert_eq!(inputs.len(), 1);
+            let input = inputs[0];
+            let (local_predicates, projections) =
+                rewrite_projection_node(expr_arena, lp_arena, &mut acc_predicates, exprs, input);
+
+            let alp = lp_arena.take(input);
+            let alp = self.push_down(alp, acc_predicates, lp_arena, expr_arena)?;
+            lp_arena.replace(input, alp);
+
+            let lp = lp.from_exprs_and_input(projections, inputs);
+            Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
+        } else {
+            let mut local_predicates = Vec::with_capacity(acc_predicates.len());
+
+            // determine new inputs by pushing down predicates
+            let new_inputs = inputs
+                .iter()
+                .map(|&node| {
+                    // first we check if we are able to push down the predicate pass this node
+                    // it could be that this node just added the column where we base the predicate on
+                    let input_schema = lp_arena.get(node).schema(lp_arena);
+                    let mut pushdown_predicates = optimizer::init_hashmap();
+                    for &predicate in acc_predicates.values() {
+                        // we can pushdown the predicate
+                        if check_input_node(predicate, input_schema, expr_arena) {
+                            let name = get_insertion_name(expr_arena, predicate, input_schema);
+                            insert_and_combine_predicate(
+                                &mut pushdown_predicates,
+                                name,
+                                predicate,
+                                expr_arena,
+                            )
+                        }
+                        // we cannot pushdown the predicate we do it here
+                        else {
+                            local_predicates.push(predicate);
+                        }
+                    }
+
+                    let alp = lp_arena.take(node);
+                    let alp = self.push_down(alp, pushdown_predicates, lp_arena, expr_arena)?;
+                    lp_arena.replace(node, alp);
+                    Ok(node)
+                })
+                .collect::<Result<Vec<_>>>()?;
+
+            let lp = lp.from_exprs_and_input(exprs, new_inputs);
+            Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
+        }
+    }
+
+    /// Filter will be done at this node, but we continue optimization
+    fn no_pushdown_restart_opt(
+        &self,
+        lp: ALogicalPlan,
+        acc_predicates: PlHashMap<Arc<str>, Node>,
+        lp_arena: &mut Arena<ALogicalPlan>,
+        expr_arena: &mut Arena<AExpr>,
+    ) -> Result<ALogicalPlan> {
+        let inputs = lp.get_inputs();
+        let exprs = lp.get_exprs();
+
+        let new_inputs = inputs
+            .iter()
+            .map(|&node| {
+                let alp = lp_arena.take(node);
+                let alp = self.push_down(alp, init_hashmap(), lp_arena, expr_arena)?;
+                lp_arena.replace(node, alp);
+                Ok(node)
+            })
+            .collect::<Result<Vec<_>>>()?;
+        let lp = lp.from_exprs_and_input(exprs, new_inputs);
+
+        // all predicates are done locally
+        let local_predicates = acc_predicates.values().copied().collect::<Vec<_>>();
+        Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
+    }
+
     /// Predicate pushdown optimizer
     ///
     /// # Arguments
@@ -72,53 +154,20 @@ impl PredicatePushDown {
     /// * `expr_arena` - The local memory arena for the expressions.
     fn push_down(
         &self,
-        logical_plan: ALogicalPlan,
-        mut acc_predicates: PlHashMap<Arc<String>, Node>,
+        lp: ALogicalPlan,
+        mut acc_predicates: PlHashMap<Arc<str>, Node>,
         lp_arena: &mut Arena<ALogicalPlan>,
         expr_arena: &mut Arena<AExpr>,
     ) -> Result<ALogicalPlan> {
         use ALogicalPlan::*;
 
-        match logical_plan {
+        match lp {
             Selection { predicate, input } => {
                 let name = roots_to_key(&aexpr_to_root_names(predicate, expr_arena));
                 insert_and_combine_predicate(&mut acc_predicates, name, predicate, expr_arena);
                 let alp = lp_arena.take(input);
                 self.push_down(alp, acc_predicates, lp_arena, expr_arena)
             }
-
-            Projection {
-                expr,
-                input,
-                schema,
-            } => {
-                for node in &expr {
-                    if is_pushdown_boundary(*node, expr_arena) {
-                        let lp = ALogicalPlanBuilder::new(input, expr_arena, lp_arena)
-                            .project(expr)
-                            .build();
-                        // do all predicates here
-                        let local_predicates = acc_predicates.into_iter().map(|(_, v)| v).collect();
-                        return Ok(self.apply_predicate(
-                            lp,
-                            local_predicates,
-                            lp_arena,
-                            expr_arena,
-                        ));
-                    }
-                }
-
-                let (local_predicates, expr) =
-                    rewrite_projection_node(expr_arena, lp_arena, &mut acc_predicates, expr, input);
-                self.pushdown_and_assign(input, acc_predicates, lp_arena, expr_arena)?;
-
-                let lp = ALogicalPlan::Projection {
-                    expr,
-                    input,
-                    schema,
-                };
-                Ok(self.apply_predicate(lp, local_predicates, lp_arena, expr_arena))
-            }
             DataFrameScan {
                 df,
                 schema,
@@ -142,9 +191,11 @@ impl PredicatePushDown {
                 schema,
             } => {
                 // predicates that will be done at this level
-                let condition = |name: Arc<String>| {
+                let condition = |name: Arc<str>| {
                     let name = &*name;
-                    name == "variable" || name == "value" || value_vars.contains(name)
+                    name == "variable"
+                        || name == "value"
+                        || value_vars.iter().any(|s| s.as_str() == name)
                 };
                 let local_predicates =
                     transfer_to_local(expr_arena, &mut acc_predicates, condition);
@@ -157,7 +208,7 @@ impl PredicatePushDown {
                     value_vars,
                     schema,
                 };
-                Ok(self.apply_predicate(lp, local_predicates, lp_arena, expr_arena))
+                Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
             }
             LocalProjection { expr, input, .. } => {
                 self.pushdown_and_assign(input, acc_predicates, lp_arena, expr_arena)?;
@@ -176,26 +227,45 @@ impl PredicatePushDown {
                     schema: Arc::new(schema),
                 })
             }
+            #[cfg(feature = "ipc")]
+            IpcScan {
+                path,
+                schema,
+                output_schema,
+                predicate,
+                aggregate,
+                options,
+            } => {
+                let predicate = predicate_at_scan(acc_predicates, predicate, expr_arena);
+
+                let lp = IpcScan {
+                    path,
+                    schema,
+                    output_schema,
+                    predicate,
+                    aggregate,
+                    options,
+                };
+                Ok(lp)
+            }
             #[cfg(feature = "parquet")]
             ParquetScan {
                 path,
                 schema,
-                with_columns,
+                output_schema,
                 predicate,
                 aggregate,
-                stop_after_n_rows,
-                cache,
+                options,
             } => {
                 let predicate = predicate_at_scan(acc_predicates, predicate, expr_arena);
 
                 let lp = ParquetScan {
                     path,
                     schema,
-                    with_columns,
+                    output_schema,
                     predicate,
                     aggregate,
-                    stop_after_n_rows,
-                    cache,
+                    options,
                 };
                 Ok(lp)
             }
@@ -203,6 +273,7 @@ impl PredicatePushDown {
             CsvScan {
                 path,
                 schema,
+                output_schema,
                 options,
                 predicate,
                 aggregate,
@@ -212,6 +283,7 @@ impl PredicatePushDown {
                 let lp = CsvScan {
                     path,
                     schema,
+                    output_schema,
                     options,
                     predicate,
                     aggregate,
@@ -219,13 +291,13 @@ impl PredicatePushDown {
                 Ok(lp)
             }
             Explode { input, columns } => {
-                let condition = |name: Arc<String>| columns.contains(&*name);
+                let condition = |name: Arc<str>| columns.iter().any(|s| s.as_str() == &*name);
                 let local_predicates =
                     transfer_to_local(expr_arena, &mut acc_predicates, condition);
 
                 self.pushdown_and_assign(input, acc_predicates, lp_arena, expr_arena)?;
                 let lp = Explode { input, columns };
-                Ok(self.apply_predicate(lp, local_predicates, lp_arena, expr_arena))
+                Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
             }
             Distinct {
                 input,
@@ -257,28 +329,7 @@ impl PredicatePushDown {
                     maintain_order,
                     subset,
                 };
-                Ok(self.apply_predicate(lp, local_predicates, lp_arena, expr_arena))
-            }
-            Aggregate {
-                input,
-                keys,
-                aggs,
-                schema,
-                apply,
-                maintain_order,
-            } => {
-                self.pushdown_and_assign(input, optimizer::init_hashmap(), lp_arena, expr_arena)?;
-
-                // dont push down predicates. An aggregation needs all rows
-                let lp = Aggregate {
-                    input,
-                    keys,
-                    aggs,
-                    schema,
-                    apply,
-                    maintain_order,
-                };
-                Ok(self.finish_at_leaf(lp, acc_predicates, lp_arena, expr_arena))
+                Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
             }
             Join {
                 input_left,
@@ -303,6 +354,7 @@ impl PredicatePushDown {
                         local_predicates.push(predicate);
                         continue;
                     }
+                    // these indicate to which tables we are going to push down the predicate
                     let mut filter_left = false;
                     let mut filter_right = false;
 
@@ -327,10 +379,20 @@ impl PredicatePushDown {
                         );
                         filter_right = true;
                     }
-                    // if not pushed down on of the tables we have to do it locally.
-                    if !(filter_left | filter_right) {
-                        local_predicates.push(predicate);
-                        continue;
+                    match (filter_left, filter_right, options.how) {
+                        // if not pushed down on of the tables we have to do it locally.
+                        (false, false, _) |
+                        // if left join and predicate only available in right table,
+                        // 'we should not filter right, because that would lead to
+                        // invalid results.
+                        // see: #2057
+                        (false, true, JoinType::Left)
+                        => {
+                            local_predicates.push(predicate);
+                            continue;
+                        },
+                        // business as usual
+                        _ => {}
                     }
                     // An outer join or left join may create null values.
                     // we also do it local
@@ -354,112 +416,32 @@ impl PredicatePushDown {
                     schema,
                     options,
                 };
-                Ok(self.apply_predicate(lp, local_predicates, lp_arena, expr_arena))
-            }
-            HStack { input, exprs, .. } => {
-                // First we get all names of added columns in this HStack operation
-                // and then we remove the predicates from the eligible container if they are
-                // dependent on data we've added in this node.
-
-                for node in &exprs {
-                    if is_pushdown_boundary(*node, expr_arena) {
-                        let lp = ALogicalPlanBuilder::new(input, expr_arena, lp_arena)
-                            .with_columns(exprs)
-                            .build();
-                        // do all predicates here
-                        let local_predicates = acc_predicates.into_iter().map(|(_, v)| v).collect();
-                        return Ok(self.apply_predicate(
-                            lp,
-                            local_predicates,
-                            lp_arena,
-                            expr_arena,
-                        ));
-                    }
-                }
-
-                let (local_predicates, exprs) = rewrite_projection_node(
-                    expr_arena,
-                    lp_arena,
-                    &mut acc_predicates,
-                    exprs,
-                    input,
-                );
-
-                self.pushdown_and_assign(input, acc_predicates, lp_arena, expr_arena)?;
-                let lp = ALogicalPlanBuilder::new(input, expr_arena, lp_arena)
-                    .with_columns(exprs)
-                    .build();
-
-                Ok(self.apply_predicate(lp, local_predicates, lp_arena, expr_arena))
+                Ok(self.optional_apply_predicate(lp, local_predicates, lp_arena, expr_arena))
             }
 
-            Udf {
-                input,
-                function,
-                predicate_pd,
-                projection_pd,
-                schema,
-            } => {
-                if predicate_pd {
-                    let input_schema = lp_arena.get(input).schema(lp_arena);
-                    let mut pushdown_predicates = optimizer::init_hashmap();
-                    let mut local_predicates = Vec::with_capacity(acc_predicates.len());
-                    for (_, predicate) in acc_predicates {
-                        if check_input_node(predicate, input_schema, expr_arena) {
-                            let name = get_insertion_name(expr_arena, predicate, input_schema);
-                            insert_and_combine_predicate(
-                                &mut pushdown_predicates,
-                                name,
-                                predicate,
-                                expr_arena,
-                            )
-                        } else {
-                            local_predicates.push(predicate);
-                        }
-                    }
-                    self.pushdown_and_assign(input, pushdown_predicates, lp_arena, expr_arena)?;
-                    let lp = Udf {
-                        input,
-                        function,
-                        predicate_pd,
-                        projection_pd,
-                        schema,
-                    };
-
-                    return Ok(self.apply_predicate(lp, local_predicates, lp_arena, expr_arena));
-                }
-                Ok(Udf {
-                    input,
-                    function,
-                    predicate_pd,
-                    projection_pd,
-                    schema,
-                })
-            }
-            lp => {
-                let inputs = lp.get_inputs();
-                let exprs = lp.get_exprs();
-
-                let new_inputs = if inputs.len() == 1 {
-                    let node = inputs[0];
-                    let alp = lp_arena.take(node);
-                    let alp = self.push_down(alp, acc_predicates, lp_arena, expr_arena)?;
-                    lp_arena.replace(node, alp);
-                    vec![node]
+            lp @ Udf { .. } => {
+                if let ALogicalPlan::Udf {
+                    predicate_pd: true, ..
+                } = lp
+                {
+                    self.pushdown_and_continue(lp, acc_predicates, lp_arena, expr_arena, false)
                 } else {
-                    inputs
-                        .iter()
-                        .map(|&node| {
-                            let alp = lp_arena.take(node);
-                            let alp =
-                                self.push_down(alp, acc_predicates.clone(), lp_arena, expr_arena)?;
-                            lp_arena.replace(node, alp);
-                            Ok(node)
-                        })
-                        .collect::<Result<Vec<_>>>()?
-                };
-
-                Ok(lp.from_exprs_and_input(exprs, new_inputs))
+                    Ok(lp)
+                }
+            }
+            // Pushed down passed these nodes
+            lp @ Cache { .. } | lp @ Union { .. } | lp @ Sort { .. } => {
+                self.pushdown_and_continue(lp, acc_predicates, lp_arena, expr_arena, false)
+            }
+            lp @ HStack {..} | lp @ Projection {..} => {
+                self.pushdown_and_continue(lp, acc_predicates, lp_arena, expr_arena, true)
+            }
+            // NOT Pushed down passed these nodes
+            // predicates influence slice sizes
+            lp @ Slice { .. }
+            // dont push down predicates. An aggregation needs all rows
+            | lp @ Aggregate {..} => {
+                self.no_pushdown_restart_opt(lp, acc_predicates, lp_arena, expr_arena)
             }
         }
     }
@@ -488,15 +470,15 @@ mod test {
         let predicate = to_aexpr(predicate_expr.clone(), &mut expr_arena);
         insert_and_combine_predicate(
             &mut acc_predicates,
-            Arc::new("foo".into()),
+            Arc::from("foo"),
             predicate,
             &mut expr_arena,
         );
-        let root = *acc_predicates.get(&String::from("foo")).unwrap();
+        let root = *acc_predicates.get("foo").unwrap();
         let expr = node_to_exp(root, &expr_arena);
         assert_eq!(
             format!("{:?}", &expr),
-            format!("{:?}", &lit(true).and(predicate_expr))
+            format!("{:?}", predicate_expr.and(lit(true)))
         );
     }
 }
