@@ -6,7 +6,7 @@ use polars_core::{datatypes::PlHashSet, prelude::*};
 fn init_vec() -> Vec<Node> {
     Vec::with_capacity(100)
 }
-fn init_set() -> PlHashSet<Arc<String>> {
+fn init_set() -> PlHashSet<Arc<str>> {
     PlHashSet::with_capacity(32)
 }
 
@@ -21,7 +21,7 @@ fn get_scan_columns(
         let mut columns = Vec::with_capacity(acc_projections.len());
         for expr in acc_projections {
             for name in aexpr_to_root_names(*expr, expr_arena) {
-                columns.push((*name).clone())
+                columns.push((*name).to_owned())
             }
         }
         with_columns = Some(columns);
@@ -38,7 +38,7 @@ fn split_acc_projections(
     acc_projections: Vec<Node>,
     down_schema: &Schema,
     expr_arena: &mut Arena<AExpr>,
-) -> (Vec<Node>, Vec<Node>, PlHashSet<Arc<String>>) {
+) -> (Vec<Node>, Vec<Node>, PlHashSet<Arc<str>>) {
     // If node above has as many columns as the projection there is nothing to pushdown.
     if down_schema.fields().len() == acc_projections.len() {
         let local_projections = acc_projections;
@@ -61,7 +61,7 @@ fn split_acc_projections(
 fn add_expr_to_accumulated(
     expr: Node,
     acc_projections: &mut Vec<Node>,
-    projected_names: &mut PlHashSet<Arc<String>>,
+    projected_names: &mut PlHashSet<Arc<str>>,
     expr_arena: &mut Arena<AExpr>,
 ) {
     for root_node in aexpr_to_root_nodes(expr, expr_arena) {
@@ -76,16 +76,31 @@ fn add_expr_to_accumulated(
 fn add_str_to_accumulated(
     name: &str,
     acc_projections: &mut Vec<Node>,
-    projected_names: &mut PlHashSet<Arc<String>>,
+    projected_names: &mut PlHashSet<Arc<str>>,
     expr_arena: &mut Arena<AExpr>,
 ) {
     // if empty: all columns are already projected.
     if !acc_projections.is_empty() {
-        let node = expr_arena.add(AExpr::Column(Arc::new(name.to_string())));
+        let node = expr_arena.add(AExpr::Column(Arc::from(name)));
         add_expr_to_accumulated(node, acc_projections, projected_names, expr_arena);
     }
 }
 
+fn update_scan_schema(
+    acc_projections: &[Node],
+    expr_arena: &Arena<AExpr>,
+    schema: &Schema,
+) -> Result<Schema> {
+    let mut new_fields = Vec::with_capacity(acc_projections.len());
+    for node in acc_projections.iter() {
+        for name in aexpr_to_root_names(*node, expr_arena) {
+            let field = schema.field_with_name(&*name)?;
+            new_fields.push(field.clone())
+        }
+    }
+    Ok(Schema::new(new_fields))
+}
+
 pub(crate) struct ProjectionPushDown {}
 
 impl ProjectionPushDown {
@@ -109,8 +124,8 @@ impl ProjectionPushDown {
         proj: Node,
         pushdown_left: &mut Vec<Node>,
         pushdown_right: &mut Vec<Node>,
-        names_left: &mut PlHashSet<Arc<String>>,
-        names_right: &mut PlHashSet<Arc<String>>,
+        names_left: &mut PlHashSet<Arc<str>>,
+        names_right: &mut PlHashSet<Arc<str>>,
         expr_arena: &mut Arena<AExpr>,
     ) -> bool {
         let mut pushed_at_least_one = false;
@@ -140,7 +155,7 @@ impl ProjectionPushDown {
         &self,
         input: Node,
         acc_projections: Vec<Node>,
-        names: PlHashSet<Arc<String>>,
+        names: PlHashSet<Arc<str>>,
         projections_seen: usize,
         lp_arena: &mut Arena<ALogicalPlan>,
         expr_arena: &mut Arena<AExpr>,
@@ -173,7 +188,7 @@ impl ProjectionPushDown {
         &self,
         logical_plan: ALogicalPlan,
         mut acc_projections: Vec<Node>,
-        mut projected_names: PlHashSet<Arc<String>>,
+        mut projected_names: PlHashSet<Arc<str>>,
         projections_seen: usize,
         lp_arena: &mut Arena<ALogicalPlan>,
         expr_arena: &mut Arena<AExpr>,
@@ -278,13 +293,14 @@ impl ProjectionPushDown {
             }
             DataFrameScan {
                 df,
-                schema,
+                mut schema,
                 selection,
                 ..
             } => {
                 let mut projection = None;
                 if !acc_projections.is_empty() {
-                    projection = Some(acc_projections)
+                    schema = Arc::new(update_scan_schema(&acc_projections, expr_arena, &*schema)?);
+                    projection = Some(acc_projections);
                 }
                 let lp = DataFrameScan {
                     df,
@@ -294,25 +310,66 @@ impl ProjectionPushDown {
                 };
                 Ok(lp)
             }
+            #[cfg(feature = "ipc")]
+            IpcScan {
+                path,
+                schema,
+                predicate,
+                aggregate,
+                mut options,
+                ..
+            } => {
+                let with_columns = get_scan_columns(&mut acc_projections, expr_arena);
+                let output_schema = if with_columns.is_none() {
+                    None
+                } else {
+                    Some(Arc::new(update_scan_schema(
+                        &acc_projections,
+                        expr_arena,
+                        &*schema,
+                    )?))
+                };
+                options.with_columns = with_columns;
+
+                let lp = IpcScan {
+                    path,
+                    schema,
+                    output_schema,
+                    predicate,
+                    aggregate,
+                    options,
+                };
+                Ok(lp)
+            }
+
             #[cfg(feature = "parquet")]
             ParquetScan {
                 path,
                 schema,
                 predicate,
                 aggregate,
-                stop_after_n_rows,
-                cache,
+                mut options,
                 ..
             } => {
                 let with_columns = get_scan_columns(&mut acc_projections, expr_arena);
+                let output_schema = if with_columns.is_none() {
+                    None
+                } else {
+                    Some(Arc::new(update_scan_schema(
+                        &acc_projections,
+                        expr_arena,
+                        &*schema,
+                    )?))
+                };
+                options.with_columns = with_columns;
+
                 let lp = ParquetScan {
                     path,
                     schema,
-                    with_columns,
+                    output_schema,
                     predicate,
                     aggregate,
-                    stop_after_n_rows,
-                    cache,
+                    options,
                 };
                 Ok(lp)
             }
@@ -326,9 +383,21 @@ impl ProjectionPushDown {
                 ..
             } => {
                 options.with_columns = get_scan_columns(&mut acc_projections, expr_arena);
+
+                let output_schema = if options.with_columns.is_none() {
+                    None
+                } else {
+                    Some(Arc::new(update_scan_schema(
+                        &acc_projections,
+                        expr_arena,
+                        &*schema,
+                    )?))
+                };
+
                 let lp = CsvScan {
                     path,
                     schema,
+                    output_schema,
                     options,
                     predicate,
                     aggregate,
@@ -494,6 +563,7 @@ impl ProjectionPushDown {
                 apply,
                 schema,
                 maintain_order,
+                dynamic_options,
             } => {
                 // the custom function may need all columns so we do the projections here.
                 if let Some(f) = apply {
@@ -504,6 +574,7 @@ impl ProjectionPushDown {
                         schema,
                         apply: Some(f),
                         maintain_order,
+                        dynamic_options,
                     };
                     let input = lp_arena.add(lp);
 
@@ -528,6 +599,13 @@ impl ProjectionPushDown {
                         add_expr_to_accumulated(*key, &mut acc_projections, &mut names, expr_arena);
                     }
 
+                    // make sure that the dynamic key is projected
+                    if let Some(options) = &dynamic_options {
+                        let node =
+                            expr_arena.add(AExpr::Column(Arc::from(options.index_column.as_str())));
+                        add_expr_to_accumulated(node, &mut acc_projections, &mut names, expr_arena);
+                    }
+
                     self.pushdown_and_assign(
                         input,
                         acc_projections,
@@ -542,6 +620,7 @@ impl ProjectionPushDown {
                         aggs,
                         apply,
                         maintain_order,
+                        dynamic_options,
                     );
                     Ok(builder.build())
                 }
@@ -627,7 +706,7 @@ impl ProjectionPushDown {
                             let root_column_name =
                                 aexpr_to_root_names(proj, expr_arena).pop().unwrap();
 
-                            let suffix = options.suffix.as_deref().unwrap_or("_right");
+                            let suffix = options.suffix.as_ref();
                             // If _right suffix exists we need to push a projection down without this
                             // suffix.
                             if root_column_name.ends_with(suffix) {
@@ -636,16 +715,16 @@ impl ProjectionPushDown {
                                     .split_at(root_column_name.len() - suffix.len());
 
                                 let downwards_name_column =
-                                    expr_arena.add(AExpr::Column(Arc::new(downwards_name.into())));
+                                    expr_arena.add(AExpr::Column(Arc::from(downwards_name)));
                                 // project downwards and locally immediately alias to prevent wrong projections
-                                if names_right.insert(Arc::new(downwards_name.to_string())) {
+                                if names_right.insert(Arc::from(downwards_name)) {
                                     pushdown_right.push(downwards_name_column);
                                 }
 
                                 // locally we project and alias
                                 let projection = expr_arena.add(AExpr::Alias(
                                     downwards_name_column,
-                                    Arc::new(format!("{}{}", downwards_name, suffix)),
+                                    Arc::from(format!("{}{}", downwards_name, suffix)),
                                 ));
                                 local_projection.push(projection);
                             }
@@ -747,41 +826,27 @@ impl ProjectionPushDown {
                     schema,
                 })
             }
-            lp @ Slice { .. } | lp @ Cache { .. } => {
+            // Slice and Cache have only inputs and exprs, so we can use same logic.
+            lp @ Slice { .. } | lp @ Cache { .. } | lp @ Union { .. } => {
                 let inputs = lp.get_inputs();
                 let exprs = lp.get_exprs();
 
-                let new_inputs = if inputs.len() == 1 {
-                    let node = inputs[0];
-                    let alp = lp_arena.take(node);
-                    let alp = self.push_down(
-                        alp,
-                        acc_projections,
-                        projected_names,
-                        projections_seen,
-                        lp_arena,
-                        expr_arena,
-                    )?;
-                    lp_arena.replace(node, alp);
-                    vec![node]
-                } else {
-                    inputs
-                        .iter()
-                        .map(|&node| {
-                            let alp = lp_arena.take(node);
-                            let alp = self.push_down(
-                                alp,
-                                acc_projections.clone(),
-                                projected_names.clone(),
-                                projections_seen,
-                                lp_arena,
-                                expr_arena,
-                            )?;
-                            lp_arena.replace(node, alp);
-                            Ok(node)
-                        })
-                        .collect::<Result<Vec<_>>>()?
-                };
+                let new_inputs = inputs
+                    .iter()
+                    .map(|&node| {
+                        let alp = lp_arena.take(node);
+                        let alp = self.push_down(
+                            alp,
+                            acc_projections.clone(),
+                            projected_names.clone(),
+                            projections_seen,
+                            lp_arena,
+                            expr_arena,
+                        )?;
+                        lp_arena.replace(node, alp);
+                        Ok(node)
+                    })
+                    .collect::<Result<Vec<_>>>()?;
 
                 Ok(lp.from_exprs_and_input(exprs, new_inputs))
             }
