@@ -1,8 +1,8 @@
 use crate::physical_plan::state::ExecutionState;
 use crate::physical_plan::PhysicalAggregation;
 use crate::prelude::*;
-use polars_arrow::array::ValueSize;
-use polars_core::chunked_array::builder::get_list_builder;
+use polars_arrow::export::arrow::{array::*, compute::concatenate::concatenate};
+use polars_arrow::prelude::QuantileInterpolOptions;
 use polars_core::frame::groupby::{fmt_groupby_column, GroupByMethod, GroupTuples};
 use polars_core::utils::NoNull;
 use polars_core::{prelude::*, POOL};
@@ -38,7 +38,7 @@ impl PhysicalExpr for AggregationExpr {
         let out = self.aggregate(df, groups, state)?.ok_or_else(|| {
             PolarsError::ComputeError("Aggregation did not return a Series".into())
         })?;
-        Ok(AggregationContext::new(out, Cow::Borrowed(groups)))
+        Ok(AggregationContext::new(out, Cow::Borrowed(groups), true))
     }
 
     fn to_field(&self, input_schema: &Schema) -> Result<Field> {
@@ -71,23 +71,23 @@ impl PhysicalAggregation for AggregationExpr {
 
         match self.agg_type {
             GroupByMethod::Min => {
-                let agg_s = ac.flat().into_owned().agg_min(ac.groups());
+                let agg_s = ac.flat_naive().into_owned().agg_min(ac.groups());
                 Ok(rename_option_series(agg_s, &new_name))
             }
             GroupByMethod::Max => {
-                let agg_s = ac.flat().into_owned().agg_max(ac.groups());
+                let agg_s = ac.flat_naive().into_owned().agg_max(ac.groups());
                 Ok(rename_option_series(agg_s, &new_name))
             }
             GroupByMethod::Median => {
-                let agg_s = ac.flat().into_owned().agg_median(ac.groups());
+                let agg_s = ac.flat_naive().into_owned().agg_median(ac.groups());
                 Ok(rename_option_series(agg_s, &new_name))
             }
             GroupByMethod::Mean => {
-                let agg_s = ac.flat().into_owned().agg_mean(ac.groups());
+                let agg_s = ac.flat_naive().into_owned().agg_mean(ac.groups());
                 Ok(rename_option_series(agg_s, &new_name))
             }
             GroupByMethod::Sum => {
-                let agg_s = ac.flat().into_owned().agg_sum(ac.groups());
+                let agg_s = ac.flat_naive().into_owned().agg_sum(ac.groups());
                 Ok(rename_option_series(agg_s, &new_name))
             }
             GroupByMethod::Count => {
@@ -97,17 +97,17 @@ impl PhysicalAggregation for AggregationExpr {
                 Ok(Some(ca.into_inner().into_series()))
             }
             GroupByMethod::First => {
-                let mut agg_s = ac.flat().into_owned().agg_first(ac.groups());
+                let mut agg_s = ac.flat_naive().into_owned().agg_first(ac.groups());
                 agg_s.rename(&new_name);
                 Ok(Some(agg_s))
             }
             GroupByMethod::Last => {
-                let mut agg_s = ac.flat().into_owned().agg_last(ac.groups());
+                let mut agg_s = ac.flat_naive().into_owned().agg_last(ac.groups());
                 agg_s.rename(&new_name);
                 Ok(Some(agg_s))
             }
             GroupByMethod::NUnique => {
-                let opt_agg = ac.flat().into_owned().agg_n_unique(ac.groups());
+                let opt_agg = ac.flat_naive().into_owned().agg_n_unique(ac.groups());
                 let opt_agg = opt_agg.map(|mut agg| {
                     agg.rename(&new_name);
                     agg.into_series()
@@ -132,14 +132,14 @@ impl PhysicalAggregation for AggregationExpr {
                 Ok(Some(column.into_series()))
             }
             GroupByMethod::Std => {
-                let agg_s = ac.flat().into_owned().agg_std(ac.groups());
+                let agg_s = ac.flat_naive().into_owned().agg_std(ac.groups());
                 Ok(rename_option_series(agg_s, &new_name))
             }
             GroupByMethod::Var => {
-                let agg_s = ac.flat().into_owned().agg_var(ac.groups());
+                let agg_s = ac.flat_naive().into_owned().agg_var(ac.groups());
                 Ok(rename_option_series(agg_s, &new_name))
             }
-            GroupByMethod::Quantile(_) => {
+            GroupByMethod::Quantile(_, _) => {
                 // implemented explicitly in AggQuantile struct
                 unimplemented!()
             }
@@ -206,26 +206,49 @@ impl PhysicalAggregation for AggregationExpr {
                 Ok(rename_option_series(agg_s, &new_name))
             }
             GroupByMethod::List => {
+                // the groups are scattered over multiple groups/sub dataframes.
+                // we now must collect them into a single group
                 let series = self.expr.evaluate(final_df, state)?;
                 let ca = series.list().unwrap();
                 let new_name = fmt_groupby_column(ca.name(), self.agg_type);
 
-                let values_type = match ca.dtype() {
-                    DataType::List(dt) => *dt.clone(),
-                    _ => unreachable!(),
-                };
+                let mut values = Vec::with_capacity(groups.len());
+                let mut can_fast_explode = true;
+
+                let mut offsets = Vec::<i64>::with_capacity(groups.len() + 1);
+                let mut length_so_far = 0i64;
+                offsets.push(length_so_far);
 
-                let mut builder =
-                    get_list_builder(&values_type, ca.get_values_size(), ca.len(), &new_name);
                 for (_, idx) in groups {
-                    // Safety
-                    // The indexes of the groupby operation are never out of bounds
-                    let ca = unsafe { ca.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
+                    let ca = unsafe {
+                        // Safety
+                        // The indexes of the groupby operation are never out of bounds
+                        ca.take_unchecked(idx.iter().map(|i| *i as usize).into())
+                    };
                     let s = ca.explode()?;
-                    builder.append_series(&s);
+                    length_so_far += s.len() as i64;
+                    offsets.push(length_so_far);
+                    values.push(s.chunks()[0].clone());
+
+                    if s.len() == 0 {
+                        can_fast_explode = false;
+                    }
                 }
-                let out = builder.finish();
-                Ok(Some(out.into_series()))
+                let vals = values.iter().map(|arr| &**arr).collect::<Vec<_>>();
+                let values: ArrayRef = concatenate(&vals).unwrap().into();
+
+                let data_type = ListArray::<i64>::default_datatype(values.data_type().clone());
+                let arr = Arc::new(ListArray::<i64>::from_data(
+                    data_type,
+                    offsets.into(),
+                    values,
+                    None,
+                )) as ArrayRef;
+                let mut ca = ListChunked::new_from_chunks(&new_name, vec![arr]);
+                if can_fast_explode {
+                    ca.set_fast_explode()
+                }
+                Ok(Some(ca.into_series()))
             }
             _ => PhysicalAggregation::aggregate(self, final_df, groups, state),
         }
@@ -240,8 +263,11 @@ impl PhysicalAggregation for AggQuantileExpr {
         state: &ExecutionState,
     ) -> Result<Option<Series>> {
         let series = self.expr.evaluate(df, state)?;
-        let new_name = fmt_groupby_column(series.name(), GroupByMethod::Quantile(self.quantile));
-        let opt_agg = series.agg_quantile(groups, self.quantile);
+        let new_name = fmt_groupby_column(
+            series.name(),
+            GroupByMethod::Quantile(self.quantile, self.interpol),
+        );
+        let opt_agg = series.agg_quantile(groups, self.quantile, self.interpol);
 
         let opt_agg = opt_agg.map(|mut agg| {
             agg.rename(&new_name);
@@ -267,11 +293,20 @@ impl PhysicalAggregation for CastExpr {
 pub struct AggQuantileExpr {
     pub(crate) expr: Arc<dyn PhysicalExpr>,
     pub(crate) quantile: f64,
+    pub(crate) interpol: QuantileInterpolOptions,
 }
 
 impl AggQuantileExpr {
-    pub fn new(expr: Arc<dyn PhysicalExpr>, quantile: f64) -> Self {
-        Self { expr, quantile }
+    pub fn new(
+        expr: Arc<dyn PhysicalExpr>,
+        quantile: f64,
+        interpol: QuantileInterpolOptions,
+    ) -> Self {
+        Self {
+            expr,
+            quantile,
+            interpol,
+        }
     }
 }
 
@@ -295,7 +330,10 @@ impl PhysicalExpr for AggQuantileExpr {
 
     fn to_field(&self, input_schema: &Schema) -> Result<Field> {
         let field = self.expr.to_field(input_schema)?;
-        let new_name = fmt_groupby_column(field.name(), GroupByMethod::Quantile(self.quantile));
+        let new_name = fmt_groupby_column(
+            field.name(),
+            GroupByMethod::Quantile(self.quantile, self.interpol),
+        );
         Ok(Field::new(&new_name, field.data_type().clone()))
     }
 
