@@ -3,7 +3,6 @@ use crate::prelude::*;
 use polars_core::frame::groupby::GroupTuples;
 use polars_core::{prelude::*, POOL};
 use rayon::prelude::*;
-use std::borrow::Cow;
 use std::sync::Arc;
 
 pub struct FilterExpr {
@@ -24,8 +23,12 @@ impl PhysicalExpr for FilterExpr {
     }
 
     fn evaluate(&self, df: &DataFrame, state: &ExecutionState) -> Result<Series> {
-        let series = self.input.evaluate(df, state)?;
-        let predicate = self.by.evaluate(df, state)?;
+        let s_f = || self.input.evaluate(df, state);
+        let predicate_f = || self.by.evaluate(df, state);
+
+        let (series, predicate) = POOL.install(|| rayon::join(s_f, predicate_f));
+        let (series, predicate) = (series?, predicate?);
+
         series.filter(predicate.bool()?)
     }
 
@@ -34,9 +37,15 @@ impl PhysicalExpr for FilterExpr {
         df: &DataFrame,
         groups: &'a GroupTuples,
         state: &ExecutionState,
-    ) -> Result<(Series, Cow<'a, GroupTuples>)> {
-        let s = self.input.evaluate(df, state)?;
-        let predicate_s = self.by.evaluate(df, state)?;
+    ) -> Result<AggregationContext<'a>> {
+        let ac_s_f = || self.input.evaluate_on_groups(df, groups, state);
+        let ac_predicate_f = || self.by.evaluate_on_groups(df, groups, state);
+
+        let (ac_s, ac_predicate) = POOL.install(|| rayon::join(ac_s_f, ac_predicate_f));
+        let (mut ac_s, ac_predicate) = (ac_s?, ac_predicate?);
+
+        let groups = ac_s.groups();
+        let predicate_s = ac_predicate.flat_naive();
         let predicate = predicate_s.bool()?;
 
         let groups = POOL.install(|| {
@@ -56,7 +65,8 @@ impl PhysicalExpr for FilterExpr {
                 .collect()
         });
 
-        Ok((s, Cow::Owned(groups)))
+        ac_s.with_groups(groups).set_original_len(false);
+        Ok(ac_s)
     }
 
     fn to_field(&self, input_schema: &Schema) -> Result<Field> {
