@@ -35,9 +35,9 @@ impl PhysicalExpr for IsNotNullExpr {
         state: &ExecutionState,
     ) -> Result<AggregationContext<'a>> {
         let mut ac = self.physical_expr.evaluate_on_groups(df, groups, state)?;
-        let s = ac.flat();
+        let s = ac.flat_naive();
         let s = s.is_not_null().into_series();
-        ac.with_series(s);
+        ac.with_series(s, false);
 
         Ok(ac)
     }
