@@ -1,13 +1,21 @@
 use super::expressions as phys_expr;
 use crate::logical_plan::Context;
+use crate::physical_plan::executors::groupby_dynamic::GroupByDynamicExec;
+#[cfg(feature = "ipc")]
+use crate::physical_plan::executors::scan::IpcExec;
+use crate::physical_plan::executors::union::UnionExec;
+use crate::prelude::shift::ShiftExpr;
 use crate::prelude::*;
-use crate::utils::{aexpr_to_root_names, aexpr_to_root_nodes, agg_source_paths, has_aexpr};
+use crate::utils::{expr_to_root_column_name, has_window_aexpr};
+use crate::{
+    logical_plan::iterator::ArenaExprIter,
+    utils::{aexpr_to_root_names, aexpr_to_root_nodes, agg_source_paths, has_aexpr},
+};
 use ahash::RandomState;
-use itertools::Itertools;
 use polars_core::prelude::*;
 use polars_core::{frame::groupby::GroupByMethod, utils::parallel_op_series};
-#[cfg(any(feature = "parquet", feature = "csv-file"))]
-use polars_io::ScanAggregation;
+#[cfg(any(feature = "parquet", feature = "csv-file", feature = "ipc"))]
+use polars_io::aggregations::ScanAggregation;
 use std::collections::HashSet;
 use std::sync::Arc;
 
@@ -22,28 +30,28 @@ fn aggregate_expr_to_scan_agg(
             let mut alias = None;
             if let AExpr::Alias(e, name) = expr_arena.get(expr) {
                 expr = *e;
-                alias = Some((**name).clone())
+                alias = Some((*name).to_string())
             };
             if let AExpr::Agg(agg) = expr_arena.get(expr) {
                 match agg {
                     AAggExpr::Min(e) => ScanAggregation::Min {
-                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).clone(),
+                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).to_string(),
                         alias,
                     },
                     AAggExpr::Max(e) => ScanAggregation::Max {
-                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).clone(),
+                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).to_string(),
                         alias,
                     },
                     AAggExpr::Sum(e) => ScanAggregation::Sum {
-                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).clone(),
+                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).to_string(),
                         alias,
                     },
                     AAggExpr::First(e) => ScanAggregation::First {
-                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).clone(),
+                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).to_string(),
                         alias,
                     },
                     AAggExpr::Last(e) => ScanAggregation::Last {
-                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).clone(),
+                        column: (*aexpr_to_root_names(*e, expr_arena).pop().unwrap()).to_string(),
                         alias,
                     },
                     _ => todo!(),
@@ -55,12 +63,8 @@ fn aggregate_expr_to_scan_agg(
         .collect()
 }
 
+#[derive(Default)]
 pub struct DefaultPlanner {}
-impl Default for DefaultPlanner {
-    fn default() -> Self {
-        Self {}
-    }
-}
 
 impl PhysicalPlanner for DefaultPlanner {
     fn create_physical_plan(
@@ -69,7 +73,7 @@ impl PhysicalPlanner for DefaultPlanner {
         lp_arena: &mut Arena<ALogicalPlan>,
         expr_arena: &mut Arena<AExpr>,
     ) -> Result<Box<dyn Executor>> {
-        self.create_initial_physical_plan(root, lp_arena, expr_arena)
+        self.create_physical_plan(root, lp_arena, expr_arena)
     }
 }
 
@@ -85,7 +89,7 @@ impl DefaultPlanner {
             .map(|e| self.create_physical_expr(*e, context, expr_arena))
             .collect()
     }
-    pub fn create_initial_physical_plan(
+    pub fn create_physical_plan(
         &self,
         root: Node,
         lp_arena: &mut Arena<ALogicalPlan>,
@@ -94,13 +98,20 @@ impl DefaultPlanner {
         use ALogicalPlan::*;
         let logical_plan = lp_arena.take(root);
         match logical_plan {
+            Union { inputs, options } => {
+                let inputs = inputs
+                    .into_iter()
+                    .map(|node| self.create_physical_plan(node, lp_arena, expr_arena))
+                    .collect::<Result<Vec<_>>>()?;
+                Ok(Box::new(UnionExec { inputs, options }))
+            }
             Melt {
                 input,
                 id_vars,
                 value_vars,
                 ..
             } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 Ok(Box::new(MeltExec {
                     input,
                     id_vars,
@@ -108,11 +119,11 @@ impl DefaultPlanner {
                 }))
             }
             Slice { input, offset, len } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 Ok(Box::new(SliceExec { input, offset, len }))
             }
             Selection { input, predicate } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let predicate =
                     self.create_physical_expr(predicate, Context::Default, expr_arena)?;
                 Ok(Box::new(FilterExec::new(predicate, input)))
@@ -121,6 +132,7 @@ impl DefaultPlanner {
             CsvScan {
                 path,
                 schema,
+                output_schema: _,
                 options,
                 predicate,
                 aggregate,
@@ -137,15 +149,36 @@ impl DefaultPlanner {
                     aggregate,
                 }))
             }
+            #[cfg(feature = "ipc")]
+            IpcScan {
+                path,
+                schema,
+                output_schema: _,
+                predicate,
+                aggregate,
+                options,
+            } => {
+                let predicate = predicate
+                    .map(|pred| self.create_physical_expr(pred, Context::Default, expr_arena))
+                    .map_or(Ok(None), |v| v.map(Some))?;
+
+                let aggregate = aggregate_expr_to_scan_agg(aggregate, expr_arena);
+                Ok(Box::new(IpcExec {
+                    path,
+                    schema,
+                    predicate,
+                    aggregate,
+                    options,
+                }))
+            }
             #[cfg(feature = "parquet")]
             ParquetScan {
                 path,
                 schema,
-                with_columns,
+                output_schema: _,
                 predicate,
                 aggregate,
-                stop_after_n_rows,
-                cache,
+                options,
             } => {
                 let predicate = predicate
                     .map(|pred| self.create_physical_expr(pred, Context::Default, expr_arena))
@@ -153,26 +186,44 @@ impl DefaultPlanner {
 
                 let aggregate = aggregate_expr_to_scan_agg(aggregate, expr_arena);
                 Ok(Box::new(ParquetExec::new(
-                    path,
-                    schema,
-                    with_columns,
-                    predicate,
-                    aggregate,
-                    stop_after_n_rows,
-                    cache,
+                    path, schema, predicate, aggregate, options,
                 )))
             }
-            Projection { expr, input, .. } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+            Projection {
+                expr,
+                input,
+                schema: _schema,
+                ..
+            } => {
+                let has_windows = expr.iter().any(|node| has_window_aexpr(*node, expr_arena));
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let phys_expr =
                     self.create_physical_expressions(&expr, Context::Default, expr_arena)?;
-                Ok(Box::new(StandardExec::new("projection", input, phys_expr)))
+                Ok(Box::new(ProjectionExec {
+                    input,
+                    expr: phys_expr,
+                    has_windows,
+                    #[cfg(test)]
+                    schema: _schema,
+                }))
             }
-            LocalProjection { expr, input, .. } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+            LocalProjection {
+                expr,
+                input,
+                schema: _schema,
+                ..
+            } => {
+                let has_windows = expr.iter().any(|node| has_window_aexpr(*node, expr_arena));
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let phys_expr =
                     self.create_physical_expressions(&expr, Context::Default, expr_arena)?;
-                Ok(Box::new(StandardExec::new("projection", input, phys_expr)))
+                Ok(Box::new(ProjectionExec {
+                    input,
+                    expr: phys_expr,
+                    has_windows,
+                    #[cfg(test)]
+                    schema: _schema,
+                }))
             }
             DataFrameScan {
                 df,
@@ -180,6 +231,14 @@ impl DefaultPlanner {
                 selection,
                 ..
             } => {
+                let has_windows = if let Some(projection) = &projection {
+                    projection
+                        .iter()
+                        .any(|node| has_window_aexpr(*node, expr_arena))
+                } else {
+                    false
+                };
+
                 let selection = selection
                     .map(|pred| self.create_physical_expr(pred, Context::Default, expr_arena))
                     .map_or(Ok(None), |v| v.map(Some))?;
@@ -188,14 +247,19 @@ impl DefaultPlanner {
                         self.create_physical_expressions(&proj, Context::Default, expr_arena)
                     })
                     .map_or(Ok(None), |v| v.map(Some))?;
-                Ok(Box::new(DataFrameExec::new(df, projection, selection)))
+                Ok(Box::new(DataFrameExec {
+                    df,
+                    projection,
+                    selection,
+                    has_windows,
+                }))
             }
             Sort {
                 input,
                 by_column,
                 reverse,
             } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let by_column =
                     self.create_physical_expressions(&by_column, Context::Default, expr_arena)?;
                 Ok(Box::new(SortExec {
@@ -205,7 +269,7 @@ impl DefaultPlanner {
                 }))
             }
             Explode { input, columns } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 Ok(Box::new(ExplodeExec { input, columns }))
             }
             Cache { input } => {
@@ -220,7 +284,7 @@ impl DefaultPlanner {
                         key.push_str(field.name())
                     }
                 }
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 Ok(Box::new(CacheExec { key, input }))
             }
             Distinct {
@@ -228,7 +292,7 @@ impl DefaultPlanner {
                 maintain_order,
                 subset,
             } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let subset = Arc::try_unwrap(subset).unwrap_or_else(|subset| (*subset).clone());
                 Ok(Box::new(DropDuplicatesExec {
                     input,
@@ -241,19 +305,57 @@ impl DefaultPlanner {
                 keys,
                 aggs,
                 apply,
-                ..
+                schema: _,
+                maintain_order,
+                dynamic_options,
             } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                #[cfg(feature = "object")]
+                let input_schema = lp_arena.get(input).schema(lp_arena).clone();
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
+
+                let mut phys_keys =
+                    self.create_physical_expressions(&keys, Context::Default, expr_arena)?;
+
+                let phys_aggs =
+                    self.create_physical_expressions(&aggs, Context::Aggregation, expr_arena)?;
+
+                if let Some(options) = dynamic_options {
+                    return Ok(Box::new(GroupByDynamicExec {
+                        input,
+                        keys: phys_keys,
+                        aggs: phys_aggs,
+                        options,
+                    }));
+                }
 
                 // We first check if we can partition the groupby on the latest moment.
                 // TODO: fix this brittle/ buggy state and implement partitioned groupby's in eager
                 let mut partitionable = true;
 
-                if keys.len() == 1 {
+                // checks:
+                //      1. complex expressions in the groupby itself are also not partitionable
+                //          in this case anything more than col("foo")
+                //      2. a custom function cannot be partitioned
+                //      3. maintain order is likely cheaper in default groupby
+                if keys.len() == 1 && apply.is_none() && !maintain_order {
+                    // complex expressions in the groupby itself are also not partitionable
+                    // in this case anything more than col("foo")
+                    if (&*expr_arena).iter(keys[0]).count() > 1 {
+                        partitionable = false;
+                    }
+
                     for agg in &aggs {
                         // make sure that we don't have a binary expr in the expr tree
-                        let matches =
-                            |e: &AExpr| matches!(e, AExpr::SortBy { .. } | AExpr::Filter { .. });
+                        let matches = |e: &AExpr| {
+                            matches!(
+                                e,
+                                AExpr::SortBy { .. }
+                                    | AExpr::Filter { .. }
+                                    | AExpr::BinaryExpr { .. }
+                                    | AExpr::BinaryFunction { .. }
+                                    | AExpr::Function { .. }
+                            )
+                        };
                         if aexpr_to_root_nodes(*agg, expr_arena).len() != 1
                             || has_aexpr(*agg, expr_arena, matches)
                         {
@@ -263,6 +365,17 @@ impl DefaultPlanner {
 
                         let agg = node_to_exp(*agg, expr_arena);
 
+                        #[cfg(feature = "object")]
+                        {
+                            let name = expr_to_root_column_name(&agg).unwrap();
+                            let fld = input_schema.field_with_name(&name).unwrap();
+
+                            if let DataType::Object(_) = fld.data_type() {
+                                partitionable = false;
+                                break;
+                            }
+                        }
+
                         // check if the aggregation type is partitionable
                         match agg {
                             Expr::Agg(AggExpr::Min(_))
@@ -283,15 +396,6 @@ impl DefaultPlanner {
                 } else {
                     partitionable = false;
                 }
-                // a custom function cannot be partitioned.
-                if apply.is_some() {
-                    partitionable = false;
-                }
-                let mut phys_keys =
-                    self.create_physical_expressions(&keys, Context::Default, expr_arena)?;
-
-                let phys_aggs =
-                    self.create_physical_expressions(&aggs, Context::Aggregation, expr_arena)?;
                 if partitionable {
                     Ok(Box::new(PartitionGroupByExec::new(
                         input,
@@ -303,23 +407,25 @@ impl DefaultPlanner {
                     )))
                 } else {
                     Ok(Box::new(GroupByExec::new(
-                        input, phys_keys, phys_aggs, apply,
+                        input,
+                        phys_keys,
+                        phys_aggs,
+                        apply,
+                        maintain_order,
                     )))
                 }
             }
             Join {
                 input_left,
                 input_right,
-                how,
                 left_on,
                 right_on,
-                allow_par,
-                force_par,
+                options,
                 ..
             } => {
-                let parallel = if force_par {
-                    force_par
-                } else if allow_par {
+                let parallel = if options.force_parallel {
+                    true
+                } else if options.allow_parallel {
                     // check if two DataFrames come from a separate source.
                     // If they don't we can parallelize,
                     // Otherwise it is in cache.
@@ -329,18 +435,13 @@ impl DefaultPlanner {
                     let mut sources_right =
                         HashSet::with_capacity_and_hasher(16, RandomState::default());
                     agg_source_paths(input_right, &mut sources_right, lp_arena);
-                    sources_left
-                        .intersection(&sources_right)
-                        .collect_vec()
-                        .is_empty()
+                    sources_left.intersection(&sources_right).next().is_none()
                 } else {
                     false
                 };
 
-                let input_left =
-                    self.create_initial_physical_plan(input_left, lp_arena, expr_arena)?;
-                let input_right =
-                    self.create_initial_physical_plan(input_right, lp_arena, expr_arena)?;
+                let input_left = self.create_physical_plan(input_left, lp_arena, expr_arena)?;
+                let input_right = self.create_physical_plan(input_right, lp_arena, expr_arena)?;
                 let left_on =
                     self.create_physical_expressions(&left_on, Context::Default, expr_arena)?;
                 let right_on =
@@ -348,22 +449,30 @@ impl DefaultPlanner {
                 Ok(Box::new(JoinExec::new(
                     input_left,
                     input_right,
-                    how,
+                    options.how,
                     left_on,
                     right_on,
                     parallel,
+                    options.suffix,
+                    options.asof_by_left,
+                    options.asof_by_right,
                 )))
             }
             HStack { input, exprs, .. } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                let has_windows = exprs.iter().any(|node| has_window_aexpr(*node, expr_arena));
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 let phys_expr =
                     self.create_physical_expressions(&exprs, Context::Default, expr_arena)?;
-                Ok(Box::new(StackExec::new(input, phys_expr)))
+                Ok(Box::new(StackExec {
+                    input,
+                    has_windows,
+                    expr: phys_expr,
+                }))
             }
             Udf {
                 input, function, ..
             } => {
-                let input = self.create_initial_physical_plan(input, lp_arena, expr_arena)?;
+                let input = self.create_physical_plan(input, lp_arena, expr_arena)?;
                 Ok(Box::new(UdfExec { input, function }))
             }
         }
@@ -382,6 +491,7 @@ impl DefaultPlanner {
                 mut function,
                 partition_by,
                 order_by: _,
+                options,
             } => {
                 // TODO! Order by
                 let group_by =
@@ -390,12 +500,25 @@ impl DefaultPlanner {
                     self.create_physical_expr(function, Context::Aggregation, expr_arena)?;
                 let mut out_name = None;
                 let mut apply_columns = aexpr_to_root_names(function, expr_arena);
-                if apply_columns.len() > 1 {
-                    return Err(PolarsError::ValueError(
-                        "Binary/Ternary function not yet supported in window expressions".into(),
-                    ));
+                // sort and then dedup removes consecutive duplicates == all duplicates
+                apply_columns.sort();
+                apply_columns.dedup();
+
+                if apply_columns.is_empty() {
+                    if has_aexpr(function, expr_arena, |e| matches!(e, AExpr::Literal(_))) {
+                        apply_columns.push(Arc::from("literal"))
+                    } else {
+                        let e = node_to_exp(function, expr_arena);
+                        return Err(PolarsError::ValueError(
+                            format!(
+                                "Cannot apply a window function, did not find a root column. \
+                            This is likely due to a syntax error in this expression: {:?}",
+                                e
+                            )
+                            .into(),
+                        ));
+                    }
                 }
-                let apply_column = apply_columns.pop().unwrap();
 
                 if let Alias(expr, name) = expr_arena.get(function) {
                     function = *expr;
@@ -405,10 +528,12 @@ impl DefaultPlanner {
 
                 Ok(Arc::new(WindowExpr {
                     group_by,
-                    apply_column,
+                    apply_columns,
                     out_name,
                     function,
                     phys_function,
+                    options,
+                    expr: node_to_exp(expression, expr_arena),
                 }))
             }
             Literal(value) => Ok(Arc::new(LiteralExpr::new(
@@ -429,22 +554,26 @@ impl DefaultPlanner {
                 column,
                 node_to_exp(expression, expr_arena),
             ))),
-            Sort { expr, reverse } => {
+            Sort { expr, options } => {
                 let phys_expr = self.create_physical_expr(expr, ctxt, expr_arena)?;
                 Ok(Arc::new(SortExpr::new(
                     phys_expr,
-                    reverse,
+                    options,
                     node_to_exp(expression, expr_arena),
                 )))
             }
             Take { expr, idx } => {
                 let phys_expr = self.create_physical_expr(expr, ctxt, expr_arena)?;
                 let phys_idx = self.create_physical_expr(idx, ctxt, expr_arena)?;
-                Ok(Arc::new(TakeExpr::new(phys_expr, phys_idx)))
+                Ok(Arc::new(TakeExpr {
+                    phys_expr,
+                    idx: phys_idx,
+                    expr: node_to_exp(expression, expr_arena),
+                }))
             }
             SortBy { expr, by, reverse } => {
                 let phys_expr = self.create_physical_expr(expr, ctxt, expr_arena)?;
-                let phys_by = self.create_physical_expr(by, ctxt, expr_arena)?;
+                let phys_by = self.create_physical_expressions(&by, ctxt, expr_arena)?;
                 Ok(Arc::new(SortByExpr::new(
                     phys_expr,
                     phys_by,
@@ -508,9 +637,9 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -530,9 +659,9 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -552,9 +681,9 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -574,9 +703,9 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -596,9 +725,9 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -614,15 +743,15 @@ impl DefaultPlanner {
                                     let s = std::mem::take(&mut s[0]);
                                     let len = s.len() as f64;
                                     parallel_op_series(|s| Ok(s.sum_as_series()), s, None)
-                                        .map(|s| s.cast::<Float64Type>().unwrap() / len)
+                                        .map(|s| s.cast(&DataType::Float64).unwrap() / len)
                                 })
                                     as Arc<dyn SeriesUdf>);
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -642,9 +771,9 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -664,9 +793,9 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -686,9 +815,9 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -700,9 +829,18 @@ impl DefaultPlanner {
                                 Ok(Arc::new(AggregationExpr::new(input, GroupByMethod::List)))
                             }
                             Context::Default => {
-                                panic!(
-                                    "list expression is only supported in the aggregation context"
-                                )
+                                let function = NoEq::new(Arc::new(move |s: &mut [Series]| {
+                                    let s = &s[0];
+                                    s.to_list().map(|ca| ca.into_series())
+                                })
+                                    as Arc<dyn SeriesUdf>);
+                                Ok(Arc::new(ApplyExpr {
+                                    inputs: vec![input],
+                                    function,
+                                    expr: node_to_exp(expression, expr_arena),
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
+                                }))
                             }
                         }
                     }
@@ -725,32 +863,36 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: Some(DataType::UInt32),
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
                     }
-                    AAggExpr::Quantile { expr, quantile } => {
+                    AAggExpr::Quantile {
+                        expr,
+                        quantile,
+                        interpol,
+                    } => {
                         // todo! add schema to get correct output type
                         let input = self.create_physical_expr(expr, ctxt, expr_arena)?;
                         match ctxt {
                             Context::Aggregation => {
-                                Ok(Arc::new(AggQuantileExpr::new(input, quantile)))
+                                Ok(Arc::new(AggQuantileExpr::new(input, quantile, interpol)))
                             }
                             Context::Default => {
                                 let function = NoEq::new(Arc::new(move |s: &mut [Series]| {
                                     let s = std::mem::take(&mut s[0]);
-                                    s.quantile_as_series(quantile)
+                                    s.quantile_as_series(quantile, interpol)
                                 })
                                     as Arc<dyn SeriesUdf>);
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: None,
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
@@ -782,18 +924,27 @@ impl DefaultPlanner {
                                 Ok(Arc::new(ApplyExpr {
                                     inputs: vec![input],
                                     function,
-                                    output_type: Some(DataType::UInt32),
                                     expr: node_to_exp(expression, expr_arena),
-                                    collect_groups: false,
+                                    collect_groups: ApplyOptions::ApplyFlat,
+                                    auto_explode: false,
                                 }))
                             }
                         }
                     }
                 }
             }
-            Cast { expr, data_type } => {
+            Cast {
+                expr,
+                data_type,
+                strict,
+            } => {
                 let phys_expr = self.create_physical_expr(expr, ctxt, expr_arena)?;
-                Ok(Arc::new(CastExpr::new(phys_expr, data_type)))
+                Ok(Arc::new(CastExpr {
+                    input: phys_expr,
+                    data_type,
+                    expr: node_to_exp(expression, expr_arena),
+                    strict,
+                }))
             }
             Ternary {
                 predicate,
@@ -813,16 +964,17 @@ impl DefaultPlanner {
             Function {
                 input,
                 function,
-                output_type,
+                output_type: _,
                 options,
             } => {
                 let input = self.create_physical_expressions(&input, ctxt, expr_arena)?;
+
                 Ok(Arc::new(ApplyExpr {
                     inputs: input,
                     function,
-                    output_type,
                     expr: node_to_exp(expression, expr_arena),
                     collect_groups: options.collect_groups,
+                    auto_explode: options.auto_explode,
                 }))
             }
             BinaryFunction {
@@ -838,20 +990,15 @@ impl DefaultPlanner {
                     input_b,
                     function,
                     output_field,
+                    expr: node_to_exp(expression, expr_arena),
                 }))
             }
             Shift { input, periods } => {
                 let input = self.create_physical_expr(input, ctxt, expr_arena)?;
-                let function = NoEq::new(Arc::new(move |s: &mut [Series]| {
-                    let s = std::mem::take(&mut s[0]);
-                    Ok(s.shift(periods))
-                }) as Arc<dyn SeriesUdf>);
-                Ok(Arc::new(ApplyExpr {
-                    inputs: vec![input],
-                    function,
-                    output_type: None,
+                Ok(Arc::new(ShiftExpr {
+                    input,
+                    periods,
                     expr: node_to_exp(expression, expr_arena),
-                    collect_groups: false,
                 }))
             }
             Slice {
@@ -864,6 +1011,7 @@ impl DefaultPlanner {
                     input,
                     offset,
                     len: length,
+                    expr: node_to_exp(expression, expr_arena),
                 }))
             }
             Reverse(expr) => {
@@ -875,9 +1023,9 @@ impl DefaultPlanner {
                 Ok(Arc::new(ApplyExpr {
                     inputs: vec![input],
                     function,
-                    output_type: None,
                     expr: node_to_exp(expression, expr_arena),
-                    collect_groups: false,
+                    collect_groups: ApplyOptions::ApplyGroups,
+                    auto_explode: false,
                 }))
             }
             Duplicated(expr) => {
@@ -889,9 +1037,9 @@ impl DefaultPlanner {
                 Ok(Arc::new(ApplyExpr {
                     inputs: vec![input],
                     function,
-                    output_type: None,
                     expr: node_to_exp(expression, expr_arena),
-                    collect_groups: false,
+                    collect_groups: ApplyOptions::ApplyGroups,
+                    auto_explode: false,
                 }))
             }
             IsUnique(expr) => {
@@ -903,9 +1051,9 @@ impl DefaultPlanner {
                 Ok(Arc::new(ApplyExpr {
                     inputs: vec![input],
                     function,
-                    output_type: None,
                     expr: node_to_exp(expression, expr_arena),
-                    collect_groups: false,
+                    collect_groups: ApplyOptions::ApplyGroups,
+                    auto_explode: false,
                 }))
             }
             Explode(expr) => {
@@ -917,9 +1065,9 @@ impl DefaultPlanner {
                 Ok(Arc::new(ApplyExpr {
                     inputs: vec![input],
                     function,
-                    output_type: None,
                     expr: node_to_exp(expression, expr_arena),
-                    collect_groups: false,
+                    collect_groups: ApplyOptions::ApplyFlat,
+                    auto_explode: false,
                 }))
             }
             Wildcard => panic!("should be no wildcard at this point"),
